""" Base class that instantiate robot control """
from abc import ABC
from covmatic_stations.station import Station, instrument_loader, labware_loader
from .robot import Robot


class RobotStationABC(Station, ABC):
    def __init__(self,
                 ot_name: str,
                 robot_manager_host: str = None,
                 robot_manager_port: int = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ot_name = ot_name
        self._robot_manager_host = robot_manager_host
        self._robot_manager_port = robot_manager_port

    @instrument_loader(0, "_robot")
    def load_robot(self):
        self._robot = Robot(robot_name=self._ot_name,
                            robot_manager_host=self._robot_manager_host,
                            robot_manager_port=self._robot_manager_port,
                            simulate=self._ctx.is_simulating())

    @instrument_loader(0, "_robot_trash")
    def load_robot_trash(self):
        self._robot_trash = Robot(robot_name="{}TRASH".format(self._ot_name),
                            robot_manager_host=self._robot_manager_host,
                            robot_manager_port=self._robot_manager_port,
                            simulate=self._ctx.is_simulating())

    def robot_pick_plate(self, slot, plate_name):
        if self._run_stage:
            self.home()
            try:
                self._robot.pick_plate(slot, plate_name)
            except Exception as e:
                self.logger.error("Error requesting pick plate {} from slot {}: {}".format(plate_name, slot, e))
                self.pause("Error in plate transfer. Please transfer manually plate {} from slot {}".format(plate_name, slot),
                           home=False)
        else:
            self.logger.info("Skipping pick plate {} from slot {} because previous stage not run.".format(plate_name, slot))

    def robot_drop_plate(self, slot, plate_name):
        if self._run_stage:
            self.home()
            try:
                self._robot.drop_plate(slot, plate_name)
            except Exception as e:
                self.logger.error("Error requesting drop plate {} to slot {}: {}".format(plate_name, slot, e))
                self.pause("Error in plate transfer. Please transfer manually plate {} to slot {}".format(plate_name, slot),
                           home=False)
        else:
            self.logger.info("Skipping pick plate {} from slot {} because previous stage not run.".format(plate_name, slot))

    def robot_trash_plate(self, pick_slot, trash_slot, plate_name="TRASH"):
        if self._run_stage:
            self.logger.info("Trashing requested from slot {} to trash slot {} for plate {}".format(pick_slot, trash_slot, plate_name))

            self.home()
            try:
                self._robot.pick_plate(pick_slot, plate_name)
                self._robot_trash.drop_plate(trash_slot, plate_name)
            except Exception as e:
                self.logger.error("Error requesting trash plate {} from slot {} to slot {}: {}".format(plate_name, pick_slot, trash_slot, e))
                self.pause("Error in plate transfer. Please move manually plate {} on slot {} to trash".format(plate_name, pick_slot),
                           home=False)
        else:
            self.logger.info("Skipping trash plate {} from slot {} to slot {} because previous stage not run.".format(
                plate_name, pick_slot, trash_slot))


import unittest
from unittest.mock import patch, MagicMock, call
from src.covmatic_robotstation.robot_station import RobotStationABC, Station
from .common import logger

TEST_SLOT = "SLOT1"
TEST_PLATE = "PLATE"


class FakeStation(RobotStationABC):
    def _tipracks(self) -> dict:
        return {}


class TestBaseClass(unittest.TestCase):
    def setUp(self) -> None:
        self._robot_patcher = patch('src.covmatic_robotstation.robot_station.Robot')
        self._mock_robot = self._robot_patcher.start()
        self._s = FakeStation(ot_name="OTTEST", logger=logger)
        self._s._ctx = MagicMock()                  # Mocking OT context calls
    def tearDown(self) -> None:
        self._robot_patcher.stop()


class TestRobotStation(TestBaseClass):
    def test_creation(self):
        self.assertTrue(self._s)

    def test_robot_is_loaded(self):
        self._s.load_robot()
        self.assertEqual(1, self._mock_robot.call_count)


class TestFunctions(TestBaseClass):
    def setUp(self) -> None:
        super().setUp()
        self._s.load_robot()

    @patch.object(Station, "home")
    def test_robot_pick_called(self, mock_home):
        self._s.robot_pick_plate(TEST_SLOT, TEST_PLATE)
        self._mock_robot().pick_plate.assert_called_once()

    @patch.object(Station, "home")
    def test_robot_pick_called_after_home(self, mock_home):
        mock_manager = MagicMock()
        mock_manager.attach_mock(mock_home, "home")
        mock_manager.attach_mock(self._mock_robot(), "robot")

        self._s.robot_pick_plate(TEST_SLOT, TEST_PLATE)

        expected_calls = [call.home(), call.robot.pick_plate(TEST_SLOT, TEST_PLATE)]

        self.assertEqual(expected_calls, mock_manager.mock_calls)

    @patch.object(Station, "home")
    def test_robot_drop_called_after_home(self, mock_home):
            mock_manager = MagicMock()
            mock_manager.attach_mock(mock_home, "home")
            mock_manager.attach_mock(self._mock_robot(), "robot")

            self._s.robot_drop_plate(TEST_SLOT, TEST_PLATE)

            expected_calls = [call.home(), call.robot.drop_plate(TEST_SLOT, TEST_PLATE)]

            self.assertEqual(expected_calls, mock_manager.mock_calls)

    @patch.object(Station, "home")
    @patch.object(Station, "pause")
    def test_if_error_ask_to_move_manually_pick(self, mock_pause, mock_home):
        self._mock_robot().pick_plate.side_effect = Exception("Unwanted exception")
        self._s.robot_pick_plate(TEST_SLOT, TEST_PLATE)
        mock_pause.assert_called()

    @patch.object(Station, "home")
    @patch.object(Station, "pause")
    def test_if_error_ask_to_move_manually_drop(self, mock_pause, mock_home):
        self._mock_robot().drop_plate.side_effect = Exception("Unwanted exception")
        self._s.robot_drop_plate(TEST_SLOT, TEST_PLATE)
        mock_pause.assert_called()